#####################################################################################################################
# Copyright IBM
#
# This PERL script was created by Shirley Dango of Rational Software.
# It was created on November 18, 2002 for the RATLC production database as part of the CC/CQ/UCM Integration project.
# The purpose is to create the Baseline record in the RATLC production database.
# The script will determine if the baseline is from a UCM environment or not.
# The user must provide the user login id, password, database name (RATLC), database set name,
# pvob, and new baseline to create.
#
# Updated by: Shirley Dango
# Updated on: April 21,2003
#	COBI16i	6/22/06		IT16:RATLC00243478 - invalid stream/pvob error
# Updated by Katie Carpenter, October & November 2007
# Updated by Stuart Poulin, Dec 2007
######################################################################################################################
# 
#This script utilized the ClearQuest COM API and the ClearCase cleartool commands.
# 
# Variable used in this script are as followings:
#	user		The login id of the ClearQuest user
#	pw		The password of the ClearQuest user
#	dbname		The schema repository connection to the database
#	dbset		The user database connection
#	pvob		The name of the project vob
#	baseline	The name of the baseline
#       projectid       The ID of the project in ClearQuest.  For example, ALM00000045
#       ucmstream       The name of the stream used in ClearCase
#	$stream_used	The name of the stream based on the pvob and baseline
#	$num_records	The number of records found in ClearQuest database
#	main		The main subroutine of the script. Added for error handling purposes.
#	get_stream	The subrountine to find the stream value based on pvob and baseline
#	get_activities	The subrountine given two baselines find the difference between them and get all the activities
#	find_contribs	The subrountine finds the contributing activites based on a set of activities
#	get_status	The subrountine finds the label status of a baseline
#	get_seed	The subrountine finds the foundation (seed) baseline based on the stream and pvob
#	get_project	The subrountine finds the project for the baseline
#	get_promotionlevel	The subrountine finds the promotion level for the baseline
#	get_components 	The subrountine finds the components of a given baseline
#	get_modcomponents	The subrountine finds the modifiable components of a baseline
#	found_baselines 	The subrountine iterates through all the foundation baselines to find the ones with modifiable components
#       total_activities  	The subrountine iterates over each foundation baseline to find the total activities
#   	create_new_record	The subroutine that creates and commits records to CQ.
#	logfile			The path where the log file will be sent
#	test			This is a flag that can be set for testing file inputs
#	help			This is a flag for the help
#	EDIT_BUILD_ACTION	This variable will set a session variable for ClearQuest access.
#######################################################################################################################
#
# Change Log:
# Flag      Item                Date        Programmer         Comments
# thl1   RATLC01087702       06/28/2005      THL               UCM Integration Scripts should use ratlperl not cqperl
# thl2   RATLC00748443       08/22/2005      THL               create_baseline_record.pl does not work in multiple pvob environment
#                                                              Changes to get_seed()
# thl3   RATLC00475437       10/18/2005      THL               the baseline create perl module needs to also check the returned 
#                                                              status from CQ API calls.
# thl4   RATLC00475598       10/19/2005      THL               SQUID Scripts for baseline and build create fail when 
#                                                              login is ldap enabled 
#
#        RATLC00396108       10/26/2007      katherns          ignore obsolete baselines
#
#        RATLC00396143       10/26/2007      katherns          change references of relCWActivities to Activities
#
#        RATLC01038670       10/26/2007      katherns          add CQ project id as an input parameter
#
#        RATLC00396084       12/18/2007      spoulin         Add -activity -from options
#        RATLC00396084       12/18/2007      spoulin         Add -nodiffbl
#        TODO		     12/27/2007      spoulin         Added composite work
#	 TODO                01/15/2008      spoulin         Added nonUCM composite support
#
#######################################################################################################################
#!ratlperl -l -w
# Changed above line to read ratlperl from cqperl thl1
#TODO fix globals
my $VERSION = "7.1.0.0";
my $INTERNALVERSION = "0.6";
use Config;
my $perl;
if ($Config{osname} =~ /win32/i) {
	$perl = 'ratlperl';
} else {
	$perl = 'cqperl';
}
use strict;
use vars qw($Bin);
use FindBin qw($Bin);
use lib "$Bin";
use Getopt::Long;
use File::Basename;
use Carp;
use FileHandle;
use CQPerlExt;
use Env;
use IO::File;
use LWP::Simple;
use POSIX;
my %seen;
my @union =();
my @isect=();
my %union =();
my %isect =();
my %count = ();
my $result;
my $stream_used;
my $oldbaseline;
my $baseline_status;
my $promotion_level;
my $mod_comps;
my @foundbaseline;
my $foundbaseline;
my $unique;
my @list;
my ($entity);
my $userLogin ; #thl4 
my $oldPVOB ; 
my $project;
my $slashvob;
my $noslashvob;

#globals added when main sub split out
my $resultSet;
my $stream_used;
my %opts;


use subs qw(main get_stream get_activities find_contribs get_status get_seed get_project get_promotionlevel get_components get_modcomponents found_baselines total_activities create_new_record);

# Unbuffer STDERR and STDOUT
select STDERR;  
$| = 1;
select STDOUT;
$| = 1;

#Define Logger parameters
use Logger qw(CONSOLE_ON CONSOLE_OFF CONSOLE_TERSE);
my $logger = Logger->new(); # the logger element

# Create the utility object
use UtilLib;
my $ut = UtilLib->new();
my $log_time = $ut->ut_file_name_date_time_stamp();

my $progname = basename $0;
my $result = main();

if ($result) {
    exit(0); # DOS Success Code
} else {
    exit(1); # DOS Error Code
}

sub main {
####
# Define the input parameters
my @opts = qw(
    user=s		
    pw=s		
    dbname=s		
    dbset=s		
    pvob=s		
    baseline=s
    projectid=s
    logfile=s
    ucmstream=s
    test	
    help
    activity=s@
    from_file=s
    diffbl!
    create_composed_of_baselines!
    create_composed_of_baselines_level=i
    add_composed_of_baseline=s@
    add_pvob_or_location=s@
);

my %opts = (
    user => undef,
    pw => undef,
    dbname => undef,
    dbset => undef,
    pvob => undef,
    baseline => undef,
    projectid => undef,
    logfile => undef,
    ucmstream => undef,      # Needed for nonUCM
    from_file => undef,
    activity => [],
    diffbl => 1,  
    create_composed_of_baselines => 1,
    create_composed_of_baselines_level => 1, #How many levels deep of add_composed_of_baseline baselines to create.  
    add_composed_of_baseline => [], # used to pass members instead of deriving from baselines.  NonUCM usage,  need to support bl@\pvob format
    add_pvob_or_location => [], # use to hold the corresponding "location" for each member baseline
);

my @required_opts = qw ( user pw dbname dbset pvob baseline projectid);

####
# Get the input parameters
unless (GetOptions(\%opts, @opts)) {
 	$logger->log_print_message("Printing Usage to the command line:");	
      	my $retval = &Usage;
      	return 0;

}

#Define the LogFile
if ($opts{logfile}) {
	my $retval;
	my $logdir = dirname($opts{logfile});
	if ( ! -e $logdir) {
		$retval = $ut->ut_create_directory_tree($logdir);
		if ( ! $retval) {
            		$logger->log_print_error(__LINE__ . ": Could not create the log file directory: $logdir");
            		$logger->log_print_error(__LINE__ . ": Script execution will continue without it.");
        	}
	}
	$retval = $logger->log_set_file_on($opts{logfile});
	if ( ! $retval) {
        	$logger->log_print_error(__LINE__ . ": Could not open the log file: $opts{logfile}.");
        	$logger->log_print_error(__LINE__ . ": Script execution will continue without it.");
   	 }
	$logger->log_print_message("Logging started at $log_time"); 
}	 

# Display the help usage
if ($opts{help}) {
	&Usage();
	return 1;
}
# Display test inputs to log file 
if ($opts{test}){
	$logger->log_print_message(" Defined user inputs:");
	$logger->log_print_message(" Defined user login is $opts{user}.");
        $logger->log_print_message(" Defined database name is $opts{dbname}.");
	$logger->log_print_message(" Defined dbset is $opts{dbset}.");
	$logger->log_print_message(" Defined Project VOB is $opts{pvob}.");
	$logger->log_print_message(" Defined Baseline name is $opts{baseline}.");	     
        $logger->log_print_message(" Defined Project ID is $opts{projectid}.");
        $logger->log_print_message(" Defined UCM Stream is $opts{ucmstream}.");
	if ($opts{logfile}){ $logger->log_print_message(" Defined logfile is $opts{logfile}.");}
	return 1;
}
####
# End program if any of the required parameters are not defined
{my  $message = "";
	for my $req (@required_opts )  {
		$message .= "Missing required option -$req \n" if ! defined $opts{$req};
	}
	if ($message ) {
		$logger->log_print_message("Missing required commandline options, Printing Usage to the command line:");
		$logger->log_print_message("$message");
		my $retval = &Usage();
		return 0;
	}
}
;

if (defined $opts{from_file} ){
	 my $fh = new FileHandle "<$opts{from_file}";
	 if (defined $fh) {
		 my $err = 0;
		 while ( my $line = $fh->getline() ) {
			 chomp $line;
			 # Allow #comment in file, strip them off.
			 $line =~ s/#.*//;  
			 next if $line =~ /^\s*$/; #ignore blank lines.
			 #Match the activity in CQ format at the beginning of each line
			 #For example "ALM0001234"  "RATLC00001234"
			 if ( $line =~ /([a-zA-Z]+\d+)/ ) {
				 push @{$opts{activity}},$1;
			 }
			 else {
				 warn "$progname : invalid activity in file $opts{from_file} : $line";
				 $err++;
			 }
		 }
		 die "Existing due to errors" if $err;
	 }
	 else {
		warn "$progname : -from_file '$opts{from_file}' : file not found or readable";
		my $retval = &Usage();
		exit 1;
	}
}
#### Start a CQ session
my $session = CQPerlExt::CQSession_Build();
# Set session variable 
#### Logon to the specified database

eval {$session->UserLogon("$opts{user}", "$opts{pw}", "$opts{dbname}", "$opts{dbset}")};
if ($@) {
$logger->log_print_error(__LINE__ . ":  Login to $opts{dbname} failed: $@"); 
return 0;
}
$userLogin = $session->GetUserLoginName() ; #thl4
$logger->log_print_success(" Login to $opts{dbname} by $userLogin succeeded."); #thl4

#### Set up vobs
##Passing in the vob to CQ requires no slashes on the vob, passing in to CC requires a slash on the vob
sub slash_noslash {
	my $pvob = shift;
	my $slashvob;
	my $noslashvob;
	if ($pvob !~ /[\/\\]/) {
	  $slashvob = '\\' .$pvob;
	  $noslashvob = $pvob;
	}
	else{
	  $slashvob = $pvob;
	  $noslashvob = $pvob ;
	  $noslashvob=~ s/\\//g; 
	}
	return ($slashvob, $noslashvob);
}
($slashvob, $noslashvob ) = &slash_noslash($opts{pvob});


#### Determine the stream used

if(!defined $opts{ucmstream}){
	$stream_used = get_stream( pvob => $slashvob, baseline => $opts{baseline} );
	if (! $stream_used){
		$logger->log_print_error(__LINE__ . ":  Couldn't determine stream used. Exiting without creating Baseline Record.");
		return 0;
	}
}
else{
	$stream_used = $opts{ucmstream};
	$logger->log_print_message(" Using the UCM Stream that was passed in: $stream_used");
}

$logger->log_print_message(" Located UCM Stream from pvob: $stream_used");
$logger->log_print_message(" The project id is:$opts{projectid}");

my @members;
if ( $opts{add_composed_of_baseline} ) {
	#Convert these pairs into {bl=>"", pvob=>""} pairs;
	$logger->log_print_message("Found -add_composed_of_baseline option ");
	for ( my $i = 0; $i < @{$opts{add_composed_of_baseline}}; $i++ ) {
		my $bl = $opts{add_composed_of_baseline}[$i];
		my $pvob = $opts{add_pvob_or_location}[$i];
		$logger->log_print_message("Baseline:$bl  pvob_or_loc:$pvob ");
		push @members, {bl=>$bl,bl_pvob=>$pvob};
	}
}

# Check if -nodiffbl or -activity  specified
if (@{$opts{activity}} or ! $opts{diffbl}){
	#activities were passed to the script on the commandline or in a file.
	#no need to diffbl in this case.
	$logger->log_print_message("-activity or -from or -nodiffbl specified: diffbl will be skipped");
	my $result = create_new_record(session => $session, stream => $stream_used, baseline => $opts{baseline}, projectid => $opts{projectid}, pvob => $opts{pvob}, user => $opts{user}, diffbl_list => \@{$opts{activity}}, members => \@members, ucm=>0 );
	if (! $result){
		$logger->log_print_error(__LINE__ . " Couldn't create baseline record.");
		return 0;
	}
}
else {

	# If composite baseline
	# get add_composed_of_baseline baselines, if no record exists for them, then create them.
	# When calling create_baseline_records.pl decrement the -create_composed_of_baselines_level by 1.
	# If this level is 0 create_baseline_records.pl should fail for composite baselines.
	# diffbl to get only local activities to add to "this" baseline record.
	#
	#### Build a query to find the baseline record to diffbl against
	# TODO add -add_composed_of_baseline option parsing for NON-UCM usage, these need to be in the form ALM0001234

	my $blcc = new get_baseline_details(baseline=>$opts{baseline},pvob=>$slashvob);
	if (! defined $blcc ) {
		$logger->log_print_error(__LINE__ . ":  unable to get baseline details baseline=>$opts{baseline},pvob=>$slashvob: $@"); 
		return 0;
	}

	my $is_composite = 0;
	# if there are add_composed_of_baseline baselines this is a composite.
	my @loc_members = $blcc->get("members");
	@members = (@members, @loc_members);
	for my $member (@members) {
		$is_composite++;
		my $member_bl = $member->{bl};
		my $member_pvob = $member->{bl_pvob};
		$logger->log_print_message(__LINE__ .  " Examining member baseline: bl = $member_bl  pvob = $member_pvob");
		my ($slashvob, $noslashvob ) = &slash_noslash($member_pvob);

		# Just a quick check for the record - without all the details. top=>1
		my $blcq = new get_baseline_record_details(session=>$session, baseline=>$member_bl, pvob=>$noslashvob,top=>1);

		if ( ! defined $blcq ) {
			#member baseline record doesn't exist;
			if ( $opts{create_composed_of_baselines} ) { # Yes create member baselines
				if ( $opts{create_composed_of_baselines_level} > 0 ) {
					my $new_level = $opts{create_composed_of_baselines_level} - 1;
					#  -create_composed_of_baselines_level $new_level
					# TODO investigate projectid
					# TODO investigate logfile naming
					my $cmd = qq|$perl $0 |;
					for my $opt (@required_opts) {
						$cmd .=  qq| -$opt "$opts{$opt}" |;
					}

					$cmd .=
					qq|$perl $0 -create_composed_of_baselines_level $new_level -create_composed_of_baselines | .
					qq|-baseline "$member_bl" | . 
					qq|-pvob "$member_pvob" | .
					qq|-logfile "$opts{logfile}.$member_bl" | .
					qq| |;
					$logger->log_print_message(__LINE__ . ": $cmd");
					my ($retval, $output) = $ut->ut_shell($cmd);
					unless ($retval ) {
						$logger->log_print_error(__LINE__ . " Error running create_baseline_record: $retval");
						$logger->log_print_error(__LINE__ . " $retval");
						return 0;
					}
				}
				else {
					# 
					$logger->log_print_error(__LINE__ . ": $member_bl $member_pvob : create_composed_of_baselines_level has reached zero, cannot create member baseline records"); 
					return 0;
				}
			}
			else {
				$logger->log_print_error(__LINE__ . ": $member_bl $member_pvob : baseline record does not exist. -create_composed_of_baselines not specified.  Cannot create member baseline records"); 
			}
		}
		else {
			#my ($member_bl) = $blcq->get("baseline");
			#my ($member_pvob) = $blcq->get("PVOB_OrLocation");
		}
	}

	my $queryDef = $session->BuildQuery("ALMBaseline");
	my $failure = $queryDef->BuildField("Name");
	$failure = $queryDef->BuildField("UCM_Stream");
	$failure = $queryDef->BuildField("history.action_timestamp"); # changed 03/13/03
	$failure = $queryDef->BuildField("PromotionLevel"); #changed 03/14/03
	$failure = $queryDef->BuildField("PVOB_OrLocation");
	$failure = $queryDef->BuildField("Obsolete");                              
	  
	my $operator = $queryDef->CreateTopNode($CQPerlExt::CQ_BOOL_OP_AND); 
	$operator->BuildFilter("PVOB_OrLocation", $CQPerlExt::CQ_COMP_OP_EQ, ["$noslashvob"]);
	$operator->BuildFilter("UCM_Stream", $CQPerlExt::CQ_COMP_OP_EQ, ["$stream_used"]);
	$operator->BuildFilter("history.action_name", $CQPerlExt::CQ_COMP_OP_EQ, ["Submit"]);
	$operator->BuildFilter("PromotionLevel", $CQPerlExt::CQ_COMP_OP_NEQ, ["REJECTED"]); # changed 03/14/03 
	$operator->BuildFilter("Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ, ["Yes"]);

	my $queryfielddefs = $queryDef->GetQueryFieldDefs(); 
	my $baselinefield = $queryfielddefs->ItemByName("history.action_timestamp");# changed 03/13/03
	$baselinefield->SetSortType($CQPerlExt::CQ_SORT_DESC); 
	$baselinefield->SetSortOrder(1); 

	eval {$resultSet = $session->BuildResultSet($queryDef)};
	# Write to log file
	if ($@) {
	$logger->log_print_error(__LINE__ . ":  Building Query Failed: $@"); 
	return 0;
	}
	else {$logger->log_print_message(" Generating query results from database");}


	my $num_records = $resultSet->ExecuteAndCountRecords(); 
	# Find the first record within ClearQuest
	for (my $i = 0; $i < $num_records; $i++){ 
		$resultSet->MoveNext;
		if ($resultSet->GetColumnValue(2) eq "$opts{baseline}") {
			$logger->log_print_error(__LINE__ . ":  Baseline record for $opts{baseline} already exists.");
			$logger->log_print_error(__LINE__ . ":  Cannot create duplicate baseline records. Exiting session.");
			CQPerlExt::CQSession_Unbuild($session);
			return 0;
	    } else {	
		$oldbaseline = $resultSet->GetColumnValue(1);
		$oldPVOB = $resultSet->GetColumnValue(5);                                  #thl2
		$logger->log_print_message(" Previous Good Baseline is $oldbaseline.");
		$i = $num_records;
	     } 
	}

	$logger->log_print_message(" The old PVOB value is: $oldPVOB");

	 
	## Determine if previous Baseline records were found in ClearQuest database

	if ($num_records eq "0"){
		$logger->log_print_message(" No previous Baseline Records found.");
		$logger->log_print_message(" Building the Baseline record with no previous baselines and UCM enabled vob.");

		## Find the project associated with the foundation baseline
		$project = get_project( stream => $stream_used, pvob => $slashvob);
		if (! $project){
			$logger->log_print_error(__LINE__ . ":  Couldn't determine stream project in pvob. 
				Exiting without creating Baseline Record.");
			return 0;
		}
		
		##Find the modifiable components of the new baseline record
		$mod_comps= get_modcomponents(project => $project, pvob => $slashvob);
		if (! $mod_comps){
			$logger->log_print_error(__LINE__ . ":  Couldn't determine modifiable components for project. 
				Exiting without creating Baseline Record.");
			return 0;
		}
		
		## Get "Seed" of Baseline record
		$oldbaseline = get_seed( stream => $stream_used, pvob => $slashvob);
		if (! $oldbaseline){
			$logger->log_print_error(__LINE__ . ":  Couldn't determine seed baseline used. 
						Exiting without creating Baseline Record.");
			return 0;
		}
		
		## Iterate through all the foundation baselines to find the ones with modifiable components
		$foundbaseline = found_baselines(oldbaseline => $oldbaseline);                            #thl2 
		if (! $foundbaseline){
			$logger->log_print_error(__LINE__ . ":  Couldn't determine seed baselines.");
			$logger->log_print_error(__LINE__ . ":  Exiting without creating Baseline Record.");
			return 0;
		}
		## Get the label status of the new baseline
		$baseline_status = get_status( pvob => $slashvob, baseline => $opts{baseline} );
		if (! $baseline_status){
			$logger->log_print_error(__LINE__ . ":  Couldn't determine baseline status. 
				Exiting without creating Baseline Record.");
			return 0;
		}
		my $diffbl_list ;                                                                         #thl1
		if ($baseline_status =~ /full/i || $baseline_status =~ /incremental/i){			
			# Iterate over each foundation baseline to find the total activities
			$diffbl_list = total_activities(foundbaseline=>$foundbaseline, %opts);            #thl2 
			}
		if (! $diffbl_list and !$is_composite) {                                                                     #thl2
		   $logger->log_print_warning(__LINE__ . ":  Couldn't determine activities for:") ;       #thl2
		   $logger->log_print_warning(__LINE__ . ":  BASELINE: $opts{baseline}");                 #thl2
		   $logger->log_print_warning(__LINE__ . ":  Exiting without creating Baseline Record."); #thl2
		   return 0 ;                                                                             #thl2
		}                                                                                         #thl2
		$logger->log_print_message(" Found all activities delivered, or this is a composite, for BASELINE: $opts{baseline}."); #thl2
		# Get rid of duplicates 
		my %diffbl_list;
		map { $diffbl_list{$_}++ } @$diffbl_list;
		my @diffbl_list = sort keys %diffbl_list;
		$logger->log_print_message(" Changed Activities: @diffbl_list");
		my $result = create_new_record(session => $session, stream => $stream_used, baseline => $opts{baseline}, projectid => $opts{projectid}, pvob => $opts{pvob}, user => $opts{user}, diffbl_list => \@diffbl_list, members=>\@members);
		if (! $result){
			$logger->log_print_error(__LINE__ . ":  Couldn't create baseline record.");
			return 0;
		}
	} # end of no previous baseline found in CQ 


	else { 
		## Previous baseline found in ClearQuest database
		$logger->log_print_message(" Building the Baseline record with previous baselines found and UCM enabled vob.");
		
		## Using the diffbl
	$oldPVOB="\\".$oldPVOB if $Config{osname} =~ /Win32/; # Windows only issue
		# If this is a composite then add nrecurse=>1
		my $diffbl_list = get_activities( BL1 => $oldbaseline, BL2 => $opts{baseline}, pvob1 => $slashvob, pvob2 => $oldPVOB, nrecurse=>$is_composite);
		if (! defined $diffbl_list and ! defined $is_composite){
			$logger->log_print_warning(__LINE__ . ":  Couldn't determine activities for:");
			$logger->log_print_warning(__LINE__ . ":  BASELINE: $oldbaseline");
			$logger->log_print_warning(__LINE__ . ": 	BASELINE: $opts{baseline}");
			$logger->log_print_warning(__LINE__ . ":  Exiting without creating Baseline Record.");
			$logger->log_print_warning(__LINE__ . " If this is the first baseline record for this project, this is the 'seed' baseline and these warnings can be ignored.");
			return 0;
		}

# Determine the number of activities 
#my $totalactvities = $@diffbl_list+1; 

		$logger->log_print_message(" Found all activities delivered between BASELINE: $oldbaseline and BASELINE: $opts{baseline}.");	
		$logger->log_print_message(" $diffbl_list");


		my $all_diffbl_list = find_contribs(pvob => $slashvob, ACT => $diffbl_list);
		$logger->log_print_message(" Found all contributing activities for rebase and deliver records");
		
		# Get rid of duplicates 
		my %diffbl_list;
		map { $diffbl_list{$_}++ } @$all_diffbl_list;
		my @diffbl_list = sort keys %diffbl_list;






		$logger->log_print_message(" Changed Activities: @diffbl_list");
		
		#execute the baseline record creation
		my $result = create_new_record(session => $session, stream => $stream_used, baseline => $opts{baseline}, projectid => $opts{projectid}, pvob => $opts{pvob}, user => $opts{user}, diffbl_list => \@diffbl_list, members => \@members);
			if (! $result){
				$logger->log_print_error(__LINE__ . " Couldn't create baseline record.");
				return 0;
		}
		
			
	} # end of $num_records loop
} # End if 

$logger->log_print_success(" Create_baseline_record script finished. Baseline Record Created.");
return 1;

} #end of sub main



####Subroutines for Program#####
# Subroutine to display the usage

sub Usage {
my $message =<<"!E!O!F!";
Usage: $progname -user <username> -pw <password> -dbname <database name> -dbset <database set> -pvob <pvob> -baseline <baseline> -projectid <project id> [-logfile <logfile>] [-ucmstream <ucmstream>] [-create_composed_of_baselines] [-test] [-help] [-activity <activity>] [-from_file <file of activities>]

$progname $VERSION :Utility to create a baseline record using the CQ API

A diffbl of the previous baseline record in ALM that uses the same pvob and ucmstream will be done.  
The activities found from this diffbl will be added to the new baseline record.   
Composite baselines with appropriate baseline dependencies must be used if multiple UCM components are used.  The script will automatically determine if the passed in baseline is a composite baseline.

Required:
 -user <user>	   The login id of the ClearQuest user.
-pw <password>	   The password of the ClearQuest user.
-dbname <dbname>   The schema repository connection to the database.
-dbset <dbset>	   The user database connection.
-pvob <\\pvob>	   The name of the project vob.
-baseline <blname> The name of the existing clearcase baseline.
-projectid <id>	   The id of the project in ALM.

Optional:
-create_composed_of_baselines	When a composite is passed to the script, the member baselines will be added to the "Composed of Baselines field".  Supplying this option will enable the automatic creation of the member baseline records if they do not already exist.  This is accomplished by recursively calling $progname.  The same Required options as supplied above will be passed to recursive calls of $progname. 
Note: Composite baselines including other composite baselines are not currently supported.  Please arrange composites to only include regular baselines.

-logfile <logfile>	Log output to supplied file along with printing to console.
-test		Just test the passed in parameters to a basic level and exit.  Do not create any records.
-help		Print this message and exit.

NonUCM options:
Use the following options to simulate UCM behavior in an nonUCM environment.
-nodiffbl	Do not try to run a diffbl (diff baseline) since this is not a UCM baseline.  
		Supply activities with the -activity or -from_file options below.
-ucmstream <stream>   Required to pass in a name for the simulated ucmstream.
-activity <activity> [-activity <activity>]	 Add given activity to baseline record created. 
-from_file <file>	Use instead of diffbl with UCM integration. 
			Given a file, with one ALM activity listed per line, 
			add these activities to the baseline record created.  
			Line Format:
			ALM00000003
			ALM00000777
			ALM00000453

NonUCM simulated composite baselines
    Composite baselines can be simulated by supplying values for the "Composed of Baselines" field.
-add_composed_of_baseline <nonUCM-baseline> 
-acc_pvob_or_location <equiv_to_pvob>
	can be abbreviated -add_comp and -add_pvob
	For each baseline you want to add to the "Composed of Baselines" field, add a pair of these options.
	Example: -add_comp existing_nonUCM_baseline_record_01 -add_pvob nonUCM_equive_to_pvob
	Both options are needed per baseline.

Note for -activity and -from_file, the passed activities must first exists in ALM.  No checking of the activity status or state is done. 

!E!O!F!

print $message;
	
return 0;
}



# Get the stream based on pvob and baseline
#
sub get_stream {
    	my %opts= @_;
        my $cmd = qq(cleartool lsbl "$opts{baseline}\@$slashvob"); 
	my ($retval, $output) = $ut->ut_shell($cmd);
   
   	unless ($retval){
		$logger->log_print_error(__LINE__ . " Stream not defined. Please check your stream or PVOB name");
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd failed to retrieve stream information.");
		$logger->log_print_error(__LINE__ . " Cleartool Output: $output");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without stream information.");
		return 0;
		my %opts = @_;
    	}
	$output =~ /stream: (.*?)@./ms;
	my $stream = $1;
	chomp $stream;
	return $stream;
}

# Find Project for the Baseline Record  
#
sub get_project {
	my %opts = @_;
#	my $cmd = qq(cleartool lsstream -fmt "%[project]p" "$opts{stream}\@$slashvob");
        my $cmd = qq(cleartool lsstream -fmt "%[project]p" "$stream_used\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        my $project;
       
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd failed to retrieve Project information");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without Project information");
		return 0;
        }
        while (defined ($project = $output)){
 		chomp $project;
		$logger->log_print_success(" Project name: $project");				
		return $project; 
    	}   	  	       	 	
} 

# Given two baselines diff them and get all the activities.
#
sub get_activities {

	my %opts = @_;
	my $nrecurse;
	$nrecurse = "-nrecurse" if $opts{nrecurse};
        my $cmd   =
      	 qq(cleartool diffbl $nrecurse -activities "$opts{BL1}\@$opts{pvob1}" "$opts{BL2}\@$opts{pvob2}"); #thl2
	my ($retval, $output) = $ut->ut_shell($cmd);
	unless ($retval) {
		$logger->log_print_warning(__LINE__ . " Cleartool command: $cmd failed to retrieve activity information.");
		$logger->log_print_warning(__LINE__ . " Baseline cannot be created without activity information.");
		$logger->log_print_warning(__LINE__ . " If this is the first baseline record for this project, this is the 'seed' baseline and these warnings can be ignored");
		return 0;
	}
	my %lact;
	my @list = split /\n/, $output;
	foreach my $act (@list) {
		chomp $act;
		if ( $act =~ />>\s([^\s]*)\s"(.*?)"/ ) {
			## SP
			my $a = $1;            # The RATLC Activity id
			my $b = $2;            # The Activity Headline
			$a =~ s/\@.*//;
                        $lact{$a} = $b;
		}
	}
	return \%lact;
}
# Given a set of activities, find contributing activities.
#
sub find_contribs {
    my %opts = @_;
    my @return;
    for my $act ( keys %{ $opts{ACT} } ) {
	next if $seen{"$act\@$slashvob"}++;
	next if $opts{ACT}->{$act} =~ /^rebase/i;
	next if $opts{ACT}->{$act} =~ /^deliver/i;
	
### SP
	my $cmd;
	if ($act !~ /\@/){
        	$cmd     = "cleartool lsact -contrib $act\@$slashvob";
	}
	else {
		$cmd     = "cleartool lsact -contrib $act";
	}
	my ($retval, $output) = $ut->ut_shell($cmd); 
        chomp $output; 
        $output =~ s/"//g;    # results enclosed in quotes "..." (?)
        if ( $output =~ /is not a deliver or rebase integration activity/ ) {
            push @return, $act;
        }
        elsif ( $output =~ /""/ ) {    # Some activities are empty?
            push @return, $act;
        }
        else {
		#Recursion
            	my @list = split /\n/,$output;
		my %list;
		for my $act ( @list ) {
			chomp $act;
			$act =~ s/"//g;
			$list{$act} = "";
	        }
            	push @return, @{ find_contribs( pvob => $slashvob, ACT => \%list) };
	}
    }
    return \@return;
}
# Find Label Status of Baseline Record  
#
sub get_status {
	my %opts = @_;
	my $cmd = qq(cleartool lsbl -fmt "%[label_status]p" "$opts{baseline}\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        my $status;
        unless ($retval){	
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd failed to retrieve label status information.");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without label status information.");
		return 0;
        }
        while (defined ($status = $output)){
 		chomp $status;
		$logger->log_print_success(" Label Status: $status");
		return $status;
    	}     	  	       	 	
}  
# Find Foundation (Seed) Baseline(s) Record 
#
sub get_seed {
        my %opts = @_;
        # thl2 we have to use -long because -fmt "%[found_bls]p" does not return the baseline selector, just its the name. 
#	my $cmd = qq(cleartool lsstream -long "$opts{stream}\@$slashvob");  
        my $cmd = qq(cleartool lsstream -long "$stream_used\@$slashvob"); 
	my ($retval, $output) = $ut->ut_shell($cmd);
        my $seed;
        my $infbls = 0 ;  #thl2
	my @fseed;
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd failed to retrieve seed record definition");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without seed record");
		return 0;
        }
        # split on the new line chars
	my @output = split(/\n/,$output) ;          #thl2
        # 
        loop: for (my $i=0;$i<=$#output;$i++) {     #thl2                                     
           $seed = $output[$i] ;                    #thl2
           chomp $seed ; 
           $seed = join(" ",split(" ",$seed)) ;     #thl2
           if ($seed =~ /^foundation baselines:/) { #thl2 Found the start of the foundation baseline section
              $infbls = 1 ;                         #thl2
              next loop ;                           #thl2
           }                                        #thl2
           elsif (($infbls) && ($seed =~ /:$/)) {   #thl2 This marks the end of the foundation baseline section
              last loop ;                           #thl2
           }                                        #thl2
           elsif ($infbls) { #thl2 still in the foundation baseline sections. Grab the bl
              $logger->log_print_success(" Foundation Baseline: $seed"); 
              push @fseed, $seed;                                        
           }
        } # end loop 
    	return \@fseed;  	       	 	
}  
# Find Components for the Baseline Record  
#
sub get_components {
	my %opts = @_;
	my $cmd = qq(cleartool lsbl -fmt "%[component]p" "$opts{baseline}\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd failed to retrieve Component information");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without Component information");
		return 0;
        }
	my @comp_list;
	my @list = split /\n/, $output;
	foreach my $comp (@list) {
		chomp $comp;
		push @comp_list, $comp;
	}
    	# Write to log file
	$logger->log_print_success(" Components for given Baseline: @comp_list");
	return \@comp_list;  	  	       	 	
} 
# Find modifiable components for the Baseline Record  
#
sub get_modcomponents {
	my %opts = @_;
	my $cmd = qq(cleartool lsproject -fmt "%[mod_comps]p" "$opts{project}\@$slashvob");
	my ($retval, $output) = $ut->ut_shell($cmd);
        unless ($retval){
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd 
		failed to retrieve Modifiable Component information");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without Modifiable Component information");
		return 0;
        }
	my @modcomp_list;
	my @list = split /\n/, $output;
	foreach my $modcomp (@list) {
		chomp $modcomp;
		push @modcomp_list, $modcomp;
	}
	$logger->log_print_success(" Modifiable Components for given Baseline: @modcomp_list");
	return \@modcomp_list;  	  	       	 	
} 

# Find Promotion Level for the Baseline Record  
#
sub get_promotionlevel { 
	my %opts = @_;
	my $cmd = qq(cleartool lsbl -l "$opts{baseline}\@$slashvob"); #021703
	my ($retval, $output) = $ut->ut_shell($cmd);
        unless (defined $retval){
		$logger->log_print_error(__LINE__ . " Cleartool command: $cmd"); 
		$logger->log_print_error(__LINE__ . " failed to retrieve Promotion Level information");
		$logger->log_print_error(__LINE__ . " Baseline cannot be created without Promotion Level information");
		return 0;
        }
	$output =~ /promotion level: (.*?)$/ms;
	my $plevel = $1;
	chomp $plevel;
	$logger->log_print_success(" Promotion level is $plevel");
	return $plevel;   	 	
} 
# Iterate through all the foundation baselines to find the ones with modifiable components

sub found_baselines{
	my %opts = @_;
	my @oldbaseline = $opts{oldbaseline};
	foreach my $item(@$oldbaseline){
                ##The format of $item is expected to be a blank delimited list with these elements   thl2
                 # baseline_selector component_selector [(modifiable)]                               thl2
                 # parse out the baseline tag and it's pvob                                          thl2
                my ($bs_selector,undef,undef) = split(/ /,$item) ;                                   #thl2
                my ($foundbaseline,$pvob) = split(/@/,$bs_selector) ;                                #thl2 
                ##Find the components of the foundation baseline record                              
                my $comps= get_components(baseline => $foundbaseline, pvob => $pvob);                #thl2
		if (! $comps){
			$logger->log_print_error(__LINE__ . " Couldn't determine component for baseline $item.");
			return 0;
		}
		
		## Determine if the component is a modifiable one		
		foreach my $e(@$mod_comps) {$union{$e}= 1}
		foreach my $e(@$comps){
		  if($union{$e}) {$isect{$e}=1}
			$union{$e} =1;
		} # end of foreach loop

		@union= keys %union;
		@isect =keys %isect;
		# Keep track of the foundation baseline that have modifiable components
		if (@isect ne ""){
		push  @foundbaseline, $bs_selector;                                                   #thl2
		} # end of tracking baselines
	} # end of iterating through all the foundation baselines
        return \@foundbaseline;
}
# Iterate over each foundation baseline to find the total activities
sub total_activities{
	my %opts = @_;
	foreach my $item(@{$opts{foundbaseline}}){	
	    if ($item ne $opts{baseline} ){
                # split out the basline name and pvob and send them to 
                # get_activities as separate opts
                my ($bn,$pv) = split(/@/,$item) ; #thl2 
		## Using the diffbl
		my $diffbl_list = get_activities( BL1 => $bn, BL2 => $opts{baseline}, pvob1 => $pv, pvob2 => $slashvob); #thl2
		if (! $diffbl_list){
			$logger->log_print_error(__LINE__ . " Couldn't determine activities for baselines");
			return 0;
		}
		$logger->log_print_message(" Found my total activities for:
				BASELINE: $item
				BASELINE: $opts{baseline}");
		$result = my $all_diffbl_list = find_contribs( ACT => $diffbl_list, pvob => $slashvob);
		if (! $result){
			$logger->log_print_error(__LINE__ . " Couldn't determine contributing activities"); 
			return 0;
		}
		$logger->log_print_message(" Found my total contributing activities");
		## Get rid of duplicates 
		my %diffbl_list;
		map { $diffbl_list{$_}++ } @$all_diffbl_list;
		my @diffbl_list = sort keys %diffbl_list;
		push (@list, @diffbl_list);
	    } # end of checking to see if oldbaseline equals new baseline
	} # end of the foreach loop

	# Final check for duplicates on all foundation baselines found
	%seen = ();
	foreach my $item (@list) {
    		$seen{$item}++;
	}
	my @uniq = keys %seen;
        return \@uniq ;
}

sub create_new_record {
	my %opts = @_;
	$opts{ucm} = 1 if (!defined $opts{ucm});
	my @diffbl_list = @{$opts{diffbl_list}};
	my ($status,$item); #thl3
	my $session = $opts{session};
	my @members = @{$opts{members}};
	# TODO if there are @members then put these on the record.
	
	## Find the promotion level of the baseline
	if ($opts{ucm}){
		$result = $promotion_level = get_promotionlevel(baseline => $opts{baseline}, pvob => $slashvob);
		if (! $result){
				$logger->log_print_error(__LINE__ . " Couldn't determine promotion level of baseline: $opts{baseline}. 
				Exiting without creating Baseline Record.");
				return 0;
			}
			
		if ($promotion_level eq "REJECTED"){
			$logger->log_print_error(__LINE__ . " BASELINE: $opts{baseline} was rejected.");
			$logger->log_print_error(__LINE__ . " Cannot create Baseline Record for Rejected Baseline");
			return 0;
		}
	}
	## Build the Baseline record
	my $entity = $session->BuildEntity("ALMBaseline");

$logger->log_print_message(" The value for the stream is: $stream_used");
$logger->log_print_message(" The value for the project id is: $opts{projectid}");
	
	#### Set field values for pvob and baseline
	my $failure = $entity->SetFieldValue("Project", "$opts{projectid}");
        $failure = $entity->SetFieldValue("PVOB_OrLocation", $noslashvob);
 	$failure = $entity->SetFieldValue("Name", "$opts{baseline}");
	$failure = $entity->SetFieldValue("PromotionLevel", "$promotion_level");
        $failure = $entity->SetFieldValue("Owner", "$userLogin");
        $failure = $entity->SetFieldValue("UCM_Stream", "$stream_used"); 

$logger->log_print_message(" Made it passed set field values");
 
        # Diffbl_list now contains the complete list of activities/contributing activities in the baseline.
	if (@diffbl_list) {
		foreach $item (@diffbl_list) {
	  		$entity->AddFieldValue("Activities",$item);
			$logger->log_print_message(" Activities added: $item");
		} # end of adding new CRs
	} # end of diffbl

	if (@members) {
		$logger->log_print_message(" Passed the member list IF");
		for my $item (@members) {
			my $bl = $item->{bl};
			my $pvob = $item->{bl_pvob};
			my ($slash, $noslash) = slash_noslash($pvob);
			$logger->log_print_message(" The value for the member is:'$noslash $bl'");
			$entity->AddFieldValue("ComposedOfBaselines", "$noslash $bl");
		} 
	} # end of members
								
	## New Baseline ID
	$logger->log_print_message(" Validating the new baseline record: $opts{baseline}");
	eval {$status = $entity->Validate()}; #thl3
	if (($@) || ($status)) {              #thl3
		$logger->log_print_error(__LINE__ . " New Baseline record has failed validation: $@");
                $logger->log_print_error(__LINE__ . " Returned Status is: $status"); #thl3
    		$entity->Revert();
   		CQPerlExt::CQSession_Unbuild($session);
    		return 0;
	}
        else {
           $logger->log_print_message(__LINE__ . " New Baseline record has validated: $@");
        }
	$status = "" ; #thl3
	$logger->log_print_message(" Committing the new baseline record: $opts{baseline}");
	eval {$status = $entity->Commit()}; #thl3
	if (($@) || ($status)) {            #thl3
		$logger->log_print_error(__LINE__ . " New Baseline record has failed to commit: $@");
                $logger->log_print_error(__LINE__ . " Returned Status is: $status"); #thl3
    		$entity->Revert();
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	$entity = undef;
	$logger->log_print_message(" The new baseline record has been created: $opts{baseline}");	
	CQPerlExt::CQSession_Unbuild($session);
	return 1;		
} 

#
# get_member_baselines
# Passed a baseline and pvob, return a list of members.  
#
sub get_member_baselines {
	my %lopts = @_; 
	my $cmd = qq{cleartool lsbl -fmt "\%[depends_on]Cp" $lopts{baseline}\@$lopts{pvob}};
	my $results = `$cmd`;
	my @members;
	if ($results ) {
		@members = split ",\s", $results;
	}
	@members;
}

#
# get_baseline_details
# Passed in a baseline and pvob, return via get method values
#

package get_baseline_details;
# Given a baseline, this package returns details on it via get("key");
sub new { 
	my $class = shift;
	my $self = {@_};
	bless($self, $class);
	if ( defined $self->{baseline} ) {
		$self->get_baseline_details();
	}
	return $self;
}

sub get_baseline_details {
	my $self = shift;
	my $cmd = qq{cleartool lsbl -l $self->{baseline}\@$self->{pvob}};
	my @lines = `$cmd`;
	# TODO test results of $cmd
	chomp @lines;
	for my $line (@lines) { #get rid of leading and trailing space in lines
		$line =~ s/^\s*//;
		$line =~ s/\s*$//;
	}
	#loop through the lines and parse out the data
	OUTER: while (my $line = shift @lines ) {
		next OUTER if $line !~ /:/;
		my ($what,$data) = split /:\s*/, $line, 2;
		if ($what =~ /change sets/ ) {
			while (my $l = shift @lines) {
				if ($l =~ /:/) { unshift @lines,$l; next OUTER; }
				push @{$self->{"activities"}},$l;
			}
		}
		if ($what =~ /depends on/ ) {
			while (my $l = shift @lines) {
				if ($l =~ /:/) { unshift @lines,$l; next OUTER; }
				#gui_11_26_2007_01@\qvmw388_pvob_01 (qvmw388_gui@\qvmw388_pvob_01)
				my ($bl,$bl_pvob,$comp,$comp_pvob) = ($l =~ /^(.*)\@(.*)\s+\((.*)\@(.*)\)/ );
				my $members = { 'bl' => $bl, 'bl_pvob' => $bl_pvob, 'comp' => $comp, 'comp_pvob' => $comp_pvob };
				push @{$self->{"_members"}},$members;
			}
		}
		else {
			push @{$self->{"_$what"}} , $data;
		}
	}
	return $self;
}

sub get {
	my $self = shift;
	my $what = shift;
	return $self->{$what} if ( exists $self->{$what} ) ;
	return @{$self->{"_$what"}} if ( exists $self->{"_$what"} ) ;
	return ();
}

#
# Given a baseline and a pvob (without slashes) search CQ for the record.
# use get("field") to get the values
# It will return an array;
# my ($act) = $bl->get("Activities");
#
package get_baseline_record_details;
 
sub new { 
	my $class = shift;
	my $self = {@_};
	bless($self, $class);
	my $results = $self->get_baseline_record_details(@_);
	return $self if $results;
	return undef;
}

sub get_baseline_record_details {
	my $self = shift;
	my %lopts = @_;
	my $queryDef = $lopts{session}->BuildQuery("ALMBaseline");
	# TODO figure out how to do real "top" top=>1

	# Get all the columns for now.
	#my $failure = $queryDef->BuildField("Name");
	my $entdef = $lopts{session}->GetEntityDef("ALMBaseline");
	my $names = $entdef->GetFieldDefNames();
	
	for my $name (@$names ) {
	   my $failure = $queryDef->BuildField($name);
	}
	  
	my $operator = $queryDef->CreateTopNode($CQPerlExt::CQ_BOOL_OP_AND); 
	$operator->BuildFilter("PVOB_OrLocation", $CQPerlExt::CQ_COMP_OP_EQ, ["$lopts{pvob}"]);
	$operator->BuildFilter("Name", $CQPerlExt::CQ_COMP_OP_EQ, ["$lopts{baseline}"]);

	my $queryfielddefs = $queryDef->GetQueryFieldDefs(); 
	#my $baselinefield = $queryfielddefs->ItemByName("history.action_timestamp");
	#$baselinefield->SetSortType($CQPerlExt::CQ_SORT_DESC); 
	#$baselinefield->SetSortOrder(1);

	my $resultSet;
	eval {$resultSet = $lopts{session}->BuildResultSet($queryDef)};
	# Write to log file
	if ($@) {
		$logger->log_print_error(__LINE__ . ":  Building Query Failed: $@"); 
		return undef;
	}
	else {
		#$logger->log_print_message(" Generating query results from database");
		1;
	}

	my %results;

	my $ct = $resultSet->ExecuteAndCountRecords(); 
	# Find the first record within ClearQuest
	return undef if ($ct == 0);

	OUTER: for (my $i = 0; $i < $ct; $i++){
	$resultSet->MoveNext;
		for (my $i = 1; $i <= $resultSet->GetNumberOfColumns(); $i++) {
			my $name = $resultSet->GetColumnLabel($i);
			#my $type = $resultSet->GetColumnType($i);
			my $value = $resultSet->GetColumnValue($i);
			$self->{$name}{$value}++;
			last OUTER if $opts{top};
			
		}
	}
	#CQPerlExt::CQSession_Unbuild($lopts{session});
	1;
}	

# Always returns an array - TODO should add "wantarray"
sub get {
	my $self = shift;
	my $what = shift;
	my @values;
	#TODO "Tie" this
	if (exists $self->{$what}) {
		if (ref $self->{$what} eq "HASH") {
			for my $k (keys %{$self->{$what}}){
				push @values, $k;
			}
		}
		elsif( ref $self->{$what} eq "ARRAY"){
			for my $k (@{$self->{$what}}) {
				push @values, $k;
			}
		}
		else {
			push @values,$self->{$what};
		}
		return @values;
	}
	else {
		return undef;
	}
} 

#### End of script
